#include "Vhub75e.h"
#include "verilated.h"
#include "verilated_vcd_c.h"
#include <array>
#include <catch2/catch_test_macros.hpp>
#include <catch2/matchers/catch_matchers.hpp>
#include <catch2/matchers/catch_matchers_all.hpp>
#include <cstdint>
#include <memory>
#include <queue>
#include <stdio.h>
#include <vector>

class CosimulatedDevice {
public:
  virtual ~CosimulatedDevice(){};

  virtual void tick() = 0;
};

// slices the RGB values for us.
uint8_t rgb_slice(uint32_t rgb, uint8_t bit) {
  if (bit > 8) {
    // todo: panic
    return 0;
  }
  uint8_t r = (rgb >> (16 + bit)) & 1;
  uint8_t g = (rgb >> (8 + bit)) & 1;
  uint8_t b = (rgb >> bit) & 1;
  return (r << 2) & (g << 1) & (b << 1);
}

class HUB75Reciever : public CosimulatedDevice {

  typedef std::vector<unsigned char> row_array;
  int xsize;
  int ysize;

  row_array row_upper{};
  row_array row_lower{};

  // the previous row values that were latched in.
  std::vector<row_array> past_rows{};
  // the pulse width for each output, in clock cycles.
  std::vector<int> pulse_widths{};

  int bit_position = 7; // the bit that is currently being shifted in

  int output_period_cnt;
  // if oe = 0, count clocks. when oe = 1, store value into
  // pulse_widths[display_bit];

  // previous latch value, used to identify when to latch.
  unsigned char prev_latch = 0;
  // previous display clock value, used to detect rising edge.
  unsigned char prev_display_clk = 0;

  unsigned char prev_oe = 0;

  // references to the panel driver signals.
  VL_IN8(&display_clk, 0, 0);
  VL_IN8(&out_enable, 0, 0);
  VL_IN8(&latch, 0, 0);
  VL_IN8(&rgb0, 2, 0);
  VL_IN8(&rgb1, 2, 0);

public:
  HUB75Reciever(int xsize, int ysize, Vhub75e &dut)
      : display_clk(dut.display_clk), out_enable(dut.out_enable),
        latch(dut.latch), rgb0(dut.panel_rgb0), rgb1(dut.panel_rgb1) {
    this->xsize = xsize;
    this->ysize = ysize;
    row_upper.clear();
  };

  // evaluates the reciever.
  virtual void tick() override {

    if (prev_display_clk == 0 && display_clk == 1) {
      // display clock rising edge.
      row_upper.push_back(rgb0);
      row_lower.push_back(rgb1);
    }

    if (prev_latch == 0 && latch == 1) {
      // latch in the data: reverse the rows, and pu
      std::reverse(row_upper.begin(), row_upper.end());
      past_rows.push_back(row_upper);
      row_upper.clear();
    }
    if (out_enable == 0) {
      if (prev_oe == 1) {
        // falling edge.
        output_period_cnt = 0;
      } else {
        output_period_cnt++;
      }
    } else {
      // rising edge: store the output
    }

    // update previous values
    prev_display_clk = display_clk;
    prev_latch = latch;
    prev_oe = out_enable;
  }

  const std::vector<row_array> &get_past_rows() { return this->past_rows; }
  const std::vector<int> &get_pulse_widths() { return this->pulse_widths; }

  // return the RGB version.
  std::vector<unsigned int> transpose() {
    auto vec = std::vector<unsigned int>();

    for (auto bits : this->row_upper) {
    }

    return vec;
  }
};

class FakeBRAM : public CosimulatedDevice {
  std::array<unsigned long, 512> ram{};

  std::queue<unsigned long> addr_lookup_q;

  unsigned short &addr_in;
  unsigned long &data_out;

  unsigned char &clk;

public:
  FakeBRAM(int latency, unsigned char &clk, unsigned short &addr_in,
           unsigned long &data_out)
      : addr_lookup_q(), addr_in(addr_in), data_out(data_out), clk(clk) {
    for (int i = 0; i < latency; i++) {
      addr_lookup_q.push(0);
    }

    for (int i = 0; i < ram.size(); i++) {
      ram[i] = i;
    }
  };

  void tick() {
    // we push and pop in the same tick: this way we keep the queue the same
    // size, acting as a pipeline delay.
    addr_lookup_q.push(addr_in);

    auto addr_to_load = addr_lookup_q.front();
    addr_lookup_q.pop();
    data_out = ram.at(addr_to_load);
  }

  // TODO: allow accessing/setting the data
  std::array<uint64_t, 512> &dump() { return ram; }

  void write(unsigned short addr, unsigned long data) { ram[addr] = data; }
};

void LineDriverTest(VerilatedContext &ctx) {
  // create the hub75e driver and run some basic tests

  // we generate 512 random color values (24 bits)
  // we load them in.
  unsigned long posedge = 0; // counts positive edges.
  unsigned long simtime = 0;
  auto dut = std::make_unique<Vhub75e>(&ctx, "dut");

  VerilatedVcdC *m_trace = new VerilatedVcdC;
  dut->trace(m_trace, 5);

  auto bram = FakeBRAM(1, dut->clk, dut->pixbuf_addr, dut->pixbuf_data);

  auto hub75 = HUB75Reciever(128, 64, *dut);
  printf("Performing basic test\n");
  bool done = false;
  m_trace->open("waveform.vcd");

  while (!done) {
    dut->clk ^= 1; // toggle clock
    dut->eval();

    if (dut->clk == 1) {
      // rising edge.
      posedge++;
      bram.tick();
    }

    dut->write_trig = posedge < 2 ? 1 : 0;

    if (posedge >= 250000) {
      done = true;
    }
    hub75.tick();
    m_trace->dump(simtime);
    simtime++;
  }
  m_trace->close();
};

// int main() {
//   auto ctx = std::make_unique<VerilatedContext>();
//   ctx->traceEverOn(true);
//   printf("hello world!\n");
//   LineDriverTest(*ctx);
// };

// test fixture to reduce amount of runtime code.
// Supports:
// adding external modules
// running the test
// storing if the done flag was raised (or not)
//
// TODO: tracing.
template <typename DUT> class VerilatorTestFixture {
  // we call .tick() on all of these. They are bound externally to the DUT.
  std::vector<std::shared_ptr<CosimulatedDevice>> external_devices;
  std::unique_ptr<VerilatedContext> ctx;
  std::unique_ptr<DUT> dut;

  unsigned long timeout =
      1000000; // clock cycles to execute before ending the simulation.
  unsigned long simtime = 0;
  unsigned long posedge = 0;

  // stores the termination-condition for the simulation.
  // if false, it means we timed out. If true, our done_condition returned true.
  bool done_condition = false;

  typedef std::function<bool(DUT &, unsigned long)> done_callback;
  // callback function to determine execution completion.
  // This can be used to add arbitrary finish-conditions to the execution.
  // If the function returns true, we set the done_set boolean to true;
  done_callback done_check;

public:
  // Create a new test fixture with a given timeout. Everything else (including
  // done-condition) can be set later.
  VerilatorTestFixture() {
    ctx = std::make_unique<VerilatedContext>();
    dut = std::make_unique<DUT>(ctx.get(), "dut");
  }

  void set_timeout(unsigned long new_timeout) { this->timeout = new_timeout; }

  void set_done_callback(done_callback d) { this->done_check = d; }

  void add_module(std::shared_ptr<CosimulatedDevice> device) {
    external_devices.push_back(device);
  }

  void exec() {
    bool done = false;
    while (!done) {
      dut->clk ^= 1;
      dut->eval();

      if (dut->clk == 1) {
        posedge++;
      }
      if (done_check) {
        if (done_check(dut, posedge)) {
          done_condition = true;
          done = true;
        }
      }

      if (posedge >= timeout) {
        done = true;
      }
      // run our external devices
      for (auto dev : external_devices) {
        dev->tick();
      }
      dut->eval(); // allow combinational logic to settle if it's being set on the negative clock.
      simtime++;
    }
  }

  // return a reference to the DUT itself. Useful for bespoke tests.
  DUT &get() { return dut; }
};

TEST_CASE("Hub75 Test") {
  auto ctx = std::make_unique<VerilatedContext>();
  // setup DUT

  unsigned long posedge = 0; // counts positive edges.
  unsigned long simtime = 0;
  auto dut = std::make_unique<Vhub75e>(ctx.get(), "dut");

  auto bram = FakeBRAM(1, dut->clk, dut->pixbuf_addr, dut->pixbuf_data);

  auto hub75 = HUB75Reciever(128, 64, *dut);
  bool done = false;

  bool driver_done = false;
  while (!done) {
    dut->clk ^= 1; // toggle clock
    dut->eval();

    if (dut->clk == 1) {
      // rising edge.
      posedge++;
    }

    dut->write_trig = posedge < 2 ? 1 : 0;
    if (dut->done) {
      done = true;
      driver_done = true;
    }

    if (posedge >= 250000) {
      done = true;
      driver_done = false;
    }
    hub75.tick();
    bram.tick();
    simtime++;
  }
  REQUIRE(driver_done);
  SECTION("Bit Sizing/Count") {

    CHECK(hub75.get_past_rows().size() == 8);

    auto rows = hub75.get_past_rows();

    for (int i = 0; i < rows.size(); i++) {
      auto r = rows[i];
      using Catch::Matchers::SizeIs;
      CHECK_THAT(r, SizeIs(128));
    }
  }
  SECTION("Pulse width") {}
}
